#pragma once

#include <atomic>
#include <chrono>

#include "Fiber.hpp"
#include "UnboundedBlockingMpscQueue.hpp"

template <typename T>
class Actor {
 private:
	enum State {
		Stopped,
		Retrieving,
		Running,
	};

	Runtime& runtime;

	std::atomic<State> state = {Stopped};

	UnboundedBlockingMpscQueue<T> queue;

	bool switchState(State oldState, State newState) {
		return state.compare_exchange_strong(oldState, newState, std::memory_order_acq_rel);
	}

	void setState(State newState) { state.store(newState, std::memory_order_release); }

	void actorLoop() {
		setState(Running);

		while (state.load(std::memory_order_acquire) == Running) {
			if (!switchState(Running, Retrieving)) {
				// Someone else must have set our state to Stopped
				break;
			}

			T t = queue.get([this] {
				// Prevent lost stop()
				switchState(Retrieving, Running);
			});

			receive(t);
		}
	}

 protected:
	Actor(Runtime& runtime) : runtime(runtime), queue(runtime) {}

	virtual void receive(T t) = 0;

	void stop() { setState(Stopped); }

 public:
	void start() {
		if (state.load(std::memory_order_acquire) != Stopped) return;

		Fiber* actorFiber = Fiber::from(std::bind(&Actor::actorLoop, this));
		runtime.schedule(*actorFiber);
	}

	void tell(T t) { queue.put(t); }

	size_t pendingMailboxItems() { return queue.size(); }

	bool waitUntilIdle(long timeout) {
		const auto start = std::chrono::steady_clock::now();
		const auto deadline = start + std::chrono::milliseconds(timeout);
		while (!(queue.size() == 0 && state.load(std::memory_order_acquire) == Retrieving)) {
			if (std::chrono::steady_clock::now() > deadline) {
				return false;
			}
		}

		return true;
	}
};
