#include "ContextManager.hpp"

#include <ostream>

#include "Context.hpp"
#include "Debug.hpp"
#include "Runtime.hpp"

ContextManager::ContextManager(Runtime& runtime) : MemoryManager(runtime), runtime(runtime) {
	auto newWorkerHook = [this]() {
		for (unsigned int i = 0; i < CONTEXT_MANAGER_FIRST_LAYER_QUEUE_SIZE * 2; ++i) {
			Context* context = new Context(this->runtime.dispatcher.getDispatchLoop());
			putFreeContext(context);
		}
	};
	// Note that it is important that this hook is executed *after*
	// the one of the MemoryManager superclass.
	runtime.addNewWorkerHook(newWorkerHook);
}

Context* ContextManager::getFreeContext() {
#ifdef CM_WITH_MEMORY_MANAGER
	bool malloced;
	void* memory = getMemory(&malloced);
	if (malloced) {
		return new (memory) Context(this->runtime.dispatcher.getDispatchLoop());
	}
	return static_cast<Context*>(memory);
#else
	return new Context(this->runtime.dispatcher.getDispatchLoop());
#endif
}

void ContextManager::putFreeContext(Context* context) {
#ifdef CM_WITH_MEMORY_MANAGER
	putMemory(context);
#else
	delete context;
#endif
}

void ContextManager::start() {
	Context* freeContext = getFreeContext();
	freeContext->start();
}

/**
 * Save the current context and start a new one.
 */
void ContextManager::saveAndStartNew(func_t freshContextHook) {
	Context* freeContext = getFreeContext();
	freeContext->setHook(freshContextHook);
	Context::getCurrentContext()->saveAndStart(freeContext);
}

/**
 * Discard the current context and resume a saved one.
 */
void ContextManager::discardAndResume(Context* context) {
	assert(context != nullptr);

	Context* contextToFree = Context::getCurrentContext();
	context->setHook([this, contextToFree] {
		LOGD("Freeing context " << contextToFree);
		putFreeContext(contextToFree);
	});
	contextToFree->discardAndResume(context);
}
