#pragma once

#include <atomic>
#include <cassert>
#include <climits>
#include <functional>

#include "Common.hpp"
#include "Debug.hpp"
#include "lib/adt/MpscQueue.hpp"

#define FIBER_FUN_TEMPLATE_ARG void(void*)
#define FIBER_FUN0_TEMPLATE_ARG void(void)

class Scheduler;
class Dispatcher;
class LawsScheduler;

class ALIGN_TO_CACHE_LINE Fiber : public Logger<LogSubsystem::F> {
 public:
	typedef std::function<FIBER_FUN_TEMPLATE_ARG> fiber_fun_t;
	typedef std::function<FIBER_FUN0_TEMPLATE_ARG> fiber_fun0_t;

	static const workeraffinity_t NOT_AFFINE = -1;

 private:
	const fiber_fun_t function;
	void* const arg;

	std::atomic<bool> runnable = {true};

	ALIGN_TO_CACHE_LINE std::atomic_uint referenceCounter = {1};

	workeraffinity_t* const affinity;

	Fiber* mpscNext = nullptr;

	/**
	 * A flag used to indicate where the fiber was from.
	 */
	unsigned int flag = 0;

	/**
	 * Dummy constructor. Used for example by the MpscQueue.
	 */
	Fiber() : Fiber([]() {}) {}

 protected:
	Fiber(fiber_fun_t function, void* arg, workeraffinity_t* affinity)
			: function(function), arg(arg), affinity(affinity){};

	// cppcheck-suppress uninitMemberVar
	explicit Fiber(fiber_fun0_t function, workeraffinity_t* affinity)
			: Fiber([function](UNUSED_ARG void* arg) { function(); }, nullptr, affinity) {}

	Fiber(fiber_fun_t function, void* arg) : Fiber(function, arg, nullptr){};

	// cppcheck-suppress uninitMemberVar
	explicit Fiber(fiber_fun0_t function) : Fiber(function, nullptr) {}

	virtual ~Fiber() = default;

	virtual void run() const;

 private:
	inline void setMpscNext(Fiber* next) { mpscNext = next; }

	inline Fiber* getMpscNext() { return mpscNext; }

	inline workeraffinity_t* getAffinityBuffer() const { return affinity; }

	inline bool setRunnableFalse() {
		bool res = runnable.load(std::memory_order_relaxed);
		if (!res) {
			// Fast path: 'runnable' was already set to false. No need
			// for an atomic operation.
			// Mesure: Impact of this "optimization". Is it really one?
			return res;
		}

		return runnable.exchange(false);
	}

	inline unsigned int doAtomicIncrRefCount() {
		assert(referenceCounter < UINT_MAX);
		return ++referenceCounter;
	}

	inline unsigned int doAtomicDecrRefCount() {
		assert(referenceCounter > 0);
		return --referenceCounter;
	}

	inline void setFlag(unsigned int flag) { this->flag = flag; }

	friend class adt::MpscQueue<Fiber>;
	friend class Scheduler;
	friend class Dispatcher;
	friend class LawsScheduler;

 public:
	unsigned int getFlag() const { return flag; }

	workeraffinity_t getAffinity() const {
		if (affinity == nullptr) {
			return NOT_AFFINE;
		}
		return *affinity;
	}

	void print() const;

	friend std::ostream& operator<<(std::ostream&, const Fiber&);

	static inline Fiber* from(fiber_fun_t function, void* arg) { return new Fiber(function, arg); }

	static inline Fiber* from(fiber_fun0_t function) { return new Fiber(function); }

	static inline Fiber* from(fiber_fun_t function, void* arg, workeraffinity_t* affinity) {
		return new Fiber(function, arg, affinity);
	}

	static inline Fiber* from(fiber_fun0_t function, workeraffinity_t* affinity) {
		return new Fiber(function, affinity);
	}
};
