#include "Runtime.hpp"

#include <pthread.h>
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#include <unistd.h>

#include <cstdlib>
#include <ctime>

// Non portable.
#include <sched.h>

#include "Common.hpp"
#include "ContextManager.hpp"
#include "Debug.hpp"
#include "emper-config.h"

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;
thread_local unsigned int Runtime::seed;
thread_local workerid_t Runtime::workerId;
RuntimeStrategy& Runtime::DEFAULT_STRATEGY = WsStrategy::INSTANCE;

Runtime::Runtime(workerid_t workerCount, RuntimeStrategy& strategy)
		: workerCount(workerCount),
			workerLatch(workerCount),
			strategy(strategy),
			scheduler(strategy.getScheduler(*this)),
			dispatcher(strategy.getDispatcher(*this)),
			contextManager(*(new ContextManager(*this))),
			atLeastOneWorkerIsSleeping(false) {
	threads = new pthread_t[workerCount];
	workerIds = new workerid_t[workerCount];

	std::srand(std::time(0));

	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		// Start non portable.
		cpu_set_t cpuset;
		CPU_ZERO(&cpuset);
		CPU_SET(i % nprocs, &cpuset);

		errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
		if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
		// End non portable.

		// Load the worker ID into the worker ID map.
		workerIds[i] = i;

		auto thread_function = [](void* voidWorkerId) -> void* {
			return currentRuntime->workerLoop(voidWorkerId);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, &workerIds[i]);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");
	}

#ifdef EMPER_STATS
	int res = std::atexit(&printLastRuntimeStats);
	if (res) {
		DIE_MSG("could not register printStats() with at_exit()");
	}
#endif
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	for (workerid_t i = 0; i < workerCount; ++i) {
		DBG("Runtime " << this << " is cancelling worker " << unsigned(i));
		errno = pthread_cancel(threads[i]);
		if (errno) {
			DIE_MSG_ERRNO("pthread_cancel() failed");
		}
	}
	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}
	DBG("Runtime " << this << " terminated");
}

void* Runtime::workerLoop(void* voidWorkerId) {
	workerId = *(workerid_t*)voidWorkerId;
	LOGD("Worker loop started by thread " << syscall(SYS_gettid));

	int oldType;
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	// Initialze the workers PRNG seed.
	seed = std::rand();

	for (auto f : newWorkerHooks) f();

	workerLatch.count_down_and_wait();

	contextManager.start();

	abort();
	return nullptr;
}

Fiber* Runtime::nextFiber() { return scheduler.nextFiber(); }

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_join(threads[i], nullptr);
	}
}

void Runtime::printStats() {
	auto runtimeStrategyStats = strategy.getStats();
	runtimeStrategyStats->print();
}

void Runtime::printLastRuntimeStats() {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats();
}

bool Runtime::inRuntime() { return dispatcher.isDispatchedControlFlow(); }

void Runtime::executeAndWait(std::function<void()> f) {
	if (inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	schedule(*fiber);

	fiberFinished.lock();
}
