#pragma once

#include <thread>
#include <vector>

#include "Common.hpp"
#include "Debug.hpp"
#include "Dispatcher.hpp"
#include "RuntimeStrategy.hpp"
#include "Scheduler.hpp"
#include "lib/sync/Latch.hpp"
#include "strategies/ws/WsStrategy.hpp"

class ContextManager;

class Runtime : public Logger<LogSubsystem::RUNTI> {
 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	static thread_local unsigned int seed;
	static thread_local workerid_t workerId;

	const workerid_t workerCount;

	std::vector<std::function<void(void)>> newWorkerHooks;

	Latch workerLatch;

	RuntimeStrategy& strategy;
	Scheduler& scheduler;
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;
	workerid_t* workerIds;

	void* workerLoop(void* workerId);

	std::mutex workerSleepMutex;
	std::condition_variable workerSleepConditionVariable;
	ALIGN_TO_CACHE_LINE std::atomic<bool> atLeastOneWorkerIsSleeping;

	static RuntimeStrategy& DEFAULT_STRATEGY;

	static void printLastRuntimeStats();

 protected:
	void addNewWorkerHook(std::function<void(void)> hook) { newWorkerHooks.push_back(hook); };

	inline void notifyAboutNewWork() {
		if (!atLeastOneWorkerIsSleeping.load(std::memory_order_relaxed)) return;

		std::lock_guard<std::mutex> lk(workerSleepMutex);
		workerSleepConditionVariable.notify_all();
	}

	void dispatcherLoopSleep() {
		std::unique_lock<std::mutex> lk(workerSleepMutex);
		atLeastOneWorkerIsSleeping.store(true, std::memory_order_relaxed);
		workerSleepConditionVariable.wait(lk);
		atLeastOneWorkerIsSleeping.store(false, std::memory_order_relaxed);
	}

 public:
	Runtime() : Runtime(std::thread::hardware_concurrency()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY) {}

	Runtime(RuntimeStrategy& strategy) : Runtime(std::thread::hardware_concurrency(), strategy) {}

	Runtime(workerid_t workerCount, RuntimeStrategy& strategy);

	~Runtime();

	inline void schedule(Fiber& fiber) { scheduler.schedule(fiber); }

	Fiber* nextFiber();

	// https://stackoverflow.com/a/3747462/194894
	static inline int rand() {
		seed = 214013 * seed + 2531011;
		return (seed >> 16) & 0x7FFF;
	}

	static inline workerid_t getWorkerId() { return workerId; }

	inline workerid_t getWorkerCount() const { return workerCount; }

	static inline Runtime* getRuntime() { return currentRuntime; }

	inline ContextManager& getContextManager() { return contextManager; }

	inline RuntimeStrategy& getStrategy() { return strategy; }

	void waitUntilFinished();

	void printStats();

	bool inRuntime();

	void executeAndWait(std::function<void()> f);

	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
};
