#pragma once

#include <mutex>
#include <queue>

#include "Blockable.hpp"
#include "Context.hpp"

template <typename T>
class UnboundedBlockingMpscQueue : public Blockable<LogSubsystem::U_B_MPSC_Q> {
 private:
	std::atomic<Context*> blockedContext = nullptr;

	bool tPopped;
	T t;

	std::queue<T> mpscQueue;
	std::mutex queueMutex;

	void tryToWakeupBlockedContext() {
		Context* context = blockedContext.exchange(nullptr);
		if (context) {
			unblock(context);
		}
	}

	void tryToGetElement(std::function<void(void)> postRetrieve) {
		// tPopped indicates that 't' is a popped and usable value.
		// Therefore if we tryToGet a new 't' while 't' is available the current 't'
		// would be overridden and dropped.
		assert(!tPopped);
		std::lock_guard<std::mutex> lock(queueMutex);
		if (!mpscQueue.empty()) {
			t = mpscQueue.front();
			mpscQueue.pop();
			tPopped = true;
			postRetrieve();
		}
	}

 public:
	UnboundedBlockingMpscQueue(Runtime& runtime) : Blockable(runtime) {}

	void put(T t) {
		{
			std::lock_guard<std::mutex> lock(queueMutex);
			mpscQueue.push(t);
		}

		// Micro optimization, see if there is a blocked context
		// before performing the atomic exchange operation.
		if (blockedContext.load() != nullptr) {
			tryToWakeupBlockedContext();
		}
	}

	T get(std::function<void(void)> postRetrieve) {
		tPopped = false;
		tryToGetElement(postRetrieve);

		if (!tPopped) {
			Context* context = Context::getCurrentContext();
			block([this, context, postRetrieve] {
				blockedContext = context;

				tryToGetElement(postRetrieve);
				if (tPopped) {
					tryToWakeupBlockedContext();
				}
			});
			if (!tPopped) {
				std::lock_guard<std::mutex> lock(queueMutex);
				// If 't' isn't already set, then mspcQueue.get() MUST
				// return an element. Note that a non-lineralizabe
				// queue may break this invariant.
				assert(!mpscQueue.empty());
				t = mpscQueue.front();
				mpscQueue.pop();
				postRetrieve();
			}
		}
		return t;
	}

	size_t size() {
		std::lock_guard<std::mutex> lock(queueMutex);
		return mpscQueue.size();
	}
};
