#include "LawsScheduler.hpp"

#include "Common.hpp"
#include "Debug.hpp"
#include "LawsStrategy.hpp"
#include "Runtime.hpp"
#include "emper-config.h"

#define EMPER_OVERFLOW_QUEUE

thread_local LawsScheduler::LawsMpscQueue LawsScheduler::priorityQueue;

thread_local LawsScheduler::WsQueue<LawsScheduler::QUEUE_SIZE> LawsScheduler::queue;

LawsScheduler::LawsScheduler(Runtime& runtime, LawsStrategy& lawsStrategy)
		: Scheduler(runtime), lawsStrategy(lawsStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new LawsScheduler::WsQueue<QUEUE_SIZE>*[workerCount];
	priorityQueues = new LawsScheduler::LawsMpscQueue*[workerCount];
	mainThreadQueue = &queue;

	auto newWorkerHook = [this]() {
		workerid_t workerId = Runtime::getWorkerId();
		queues[workerId] = &queue;
		priorityQueues[workerId] = &priorityQueue;
	};
	addNewWorkerHook(newWorkerHook);
}

void LawsScheduler::schedule(Fiber& fiber) {
	LOGD("Scheduling fiber " << &fiber);

	workeraffinity_t* const affinity_buffer = getAffinityBuffer(fiber);
	if (affinity_buffer) {
		workeraffinity_t affinity = *affinity_buffer;
		workerid_t workerId = Runtime::getWorkerId();
		if (affinity == workerId) {
			goto scheduleToLocalWsQueue;
		}

		if (affinity == Fiber::NOT_AFFINE) {
			goto scheduleToLocalWsQueue;
		}

		// We found a fiber to schedule on a remote prority queue.
		increaseRefCount(fiber);
		priorityQueues[affinity]->enqueue(&fiber);
#ifdef EMPER_STATS
		lawsStrategy.scheduledFibersToRemotePriority.fetch_add(1, std::memory_order_relaxed);
#endif
	}

scheduleToLocalWsQueue:
	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
#ifdef EMPER_OVERFLOW_QUEUE
		priorityQueue.enqueue(&fiber);
#else
		ABORT("Could not push fiber " << &fiber << " into queue");
#endif
	} else {
#ifdef EMPER_STATS
		lawsStrategy.scheduledFibersToLocal.fetch_add(1, std::memory_order_relaxed);
#endif
	}

#ifdef EMPER_WORKER_SLEEP
	notifyRuntimeAboutNewWork();
#endif
}

Fiber* LawsScheduler::nextFiber() {
	Fiber* fiber = priorityQueue.dequeue();
	if (fiber != nullptr) {
		// We fetched a fiber from your local priority queue.
#ifdef EMPER_STATS
		unsigned int flag = static_cast<unsigned int>(LawsStrategy::FiberSource::fromPriority);
		fiber->setFlag(flag);
#endif
		return fiber;
	}

	bool poped = queue.popBottom(&fiber);

	if (likely(poped)) {
#ifdef EMPER_STATS
		unsigned int flag = static_cast<unsigned int>(LawsStrategy::FiberSource::fromLocal);
		fiber->setFlag(flag);
#endif
		return fiber;
	}

	const workerid_t workerCount = runtime.getWorkerCount();
	workerid_t startWorkerId = Runtime::rand() % workerCount;

	if (unlikely(startWorkerId == Runtime::getWorkerId())) startWorkerId++;
	workerid_t victim = startWorkerId % workerCount;

	poped = queues[victim]->popTop(&fiber);
	if (poped) {
#ifdef EMPER_STATS
		unsigned int flag = static_cast<unsigned int>(LawsStrategy::FiberSource::stolen);
		fiber->setFlag(flag);
#endif
		return fiber;
	}

	// TODO: Reduce pressure on mainThreadQueue by only checking every N-th time.

	// Try the main thread queue to get work as last resort.
	poped = mainThreadQueue->popTop(&fiber);
	if (poped) {
#ifdef EMPER_STATS
		unsigned int flag = static_cast<unsigned int>(LawsStrategy::FiberSource::mainThread);
		fiber->setFlag(flag);
#endif
		return fiber;
	}

	return nullptr;
}
