#include "WsScheduler.hpp"

#include "Common.hpp"
#include "Debug.hpp"
#include "Runtime.hpp"
#include "emper-config.h"

thread_local WsScheduler::WsQueue<WsScheduler::QUEUE_SIZE> WsScheduler::queue;

WsScheduler::WsScheduler(Runtime& runtime, WsStrategy& wsStrategy)
		: Scheduler(runtime), wsStrategy(wsStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new WsScheduler::WsQueue<QUEUE_SIZE>*[workerCount];
	mainThreadQueue = &queue;

	auto newWorkerHook = [this]() { queues[Runtime::getWorkerId()] = &queue; };
	addNewWorkerHook(newWorkerHook);
}

void WsScheduler::schedule(Fiber& fiber) {
	LOGD("Scheduling fiber " << &fiber);

	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
		// Work-stealing should not use an overflow queue
		// (EMPER_OVERFLOW_QUEUE), because of the extra overhead
		// required to check that queue for work, so we have to abort
		// here.
		ABORT("Could not push fiber " << &fiber << " into queue");
	}

#ifdef EMPER_STATS
	wsStrategy.scheduledFibers.fetch_add(1, std::memory_order_relaxed);
#endif

#ifdef EMPER_WORKER_SLEEP
	notifyRuntimeAboutNewWork();
#endif
}

Fiber* WsScheduler::nextFiber() {
	Fiber* fiber;
	bool poped = queue.popBottom(&fiber);

	if (likely(poped)) {
#ifdef EMPER_STATS
		wsStrategy.nextFiberFromLocal.fetch_add(1, std::memory_order_relaxed);
#endif
		return fiber;
	}

	const workerid_t workerCount = runtime.getWorkerCount();
	workerid_t startWorkerId = Runtime::rand() % workerCount;
	// TODO: See how reducing the loop bound affects things.
	for (workerid_t i = 0; i < workerCount; ++i) {
		workerid_t victim = (startWorkerId + i) % workerCount;

		// Don't steal from ourselves.
		if (unlikely(victim == Runtime::getWorkerId())) continue;

		poped = queues[victim]->popTop(&fiber);
		if (poped) {
#ifdef EMPER_STATS
			wsStrategy.nextFiberStolen.fetch_add(1, std::memory_order_relaxed);
#endif
			return fiber;
		}
	}

	// Try the main thread queue to get work as last resort.
	poped = mainThreadQueue->popTop(&fiber);
	if (poped) return fiber;

	return nullptr;
}
