#!/usr/bin/env bash

set -e

# shellcheck disable=SC1090
. "$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )/setup.sh"

printVerbose() {
	if ! $VERBOSE; then
		return;
	fi
	echo "$@"
}

checkVersionHeader() {
	local version="${1}"

	mapfile -t VERSION_COMPONENTS < <(echo "${version}" | grep --extended-regexp --only-matching '[^\.-]+' -)

	EMPER_MAJOR_VERSION="${VERSION_COMPONENTS[0]}"
	EMPER_MINOR_VERSION="${VERSION_COMPONENTS[1]}"
	EMPER_PATCH_VERSION="${VERSION_COMPONENTS[2]}"
	EMPER_EXTRA_VERSION=""
	EMPER_IS_SNAPSHOT=false

	local numComponents=${#VERSION_COMPONENTS[@]}

	if [[ ${numComponents} -ge 4 ]]; then
		EMPER_EXTRA_VERSION="${VERSION_COMPONENTS[3]}"
	fi

	if [[ ${numComponents} -eq 5 ]]; then
		if [[ "${VERSION_COMPONENTS[4]}" != SNAPSHOT ]]; then
			echo "Invalid snapshot identifier: ${VERSION_COMPONENTS[4]}"
			exit 2
		fi

		EMPER_IS_SNAPSHOT=true
	fi

	if [[ -z $EMPER_MAJOR_VERSION ]]; then
		echo "EMPER major version not set"
		exit 1
	fi
	if [[ -z $EMPER_MINOR_VERSION ]]; then
		echo "EMPER minor version not set"
		exit 1
	fi
	if [[ -z $EMPER_PATCH_VERSION ]]; then
		echo "EMPER patch version not set"
		exit 1
	fi
	

	pushd . > /dev/null
	cd "${BASEDIR}"
	EMPER_GIT_COMMIT=$(git describe --always --tags --dirty=+)
	popd > /dev/null

	EMPER_SHORT_VERSION="${EMPER_MAJOR_VERSION}.${EMPER_MINOR_VERSION}.${EMPER_PATCH_VERSION}"
	if [[ -n ${EMPER_EXTRA_VERSION} ]]; then
		EMPER_SHORT_VERSION+="-${EMPER_EXTRA_VERSION}"
	fi
	if $EMPER_IS_SNAPSHOT; then
		EMPER_SHORT_VERSION+="-SNAPSHOT"
	fi

	EMPER_FULL_VERSION="${EMPER_SHORT_VERSION} (${EMPER_GIT_COMMIT})"

	if [[ -f "$versionHeader" ]] && grep -q "${EMPER_FULL_VERSION}" "$versionHeader"; then
		printVerbose "Version file up-to-date. EMPER version ${EMPER_SHORT_VERSION}"
		exit 0
	fi

	cat <<EOF > "$versionHeader"
#pragma once

#define EMPER_MAJOR_VERSION ${EMPER_MAJOR_VERSION}
#define EMPER_MINOR_VERSION ${EMPER_MINOR_VERSION}
#define EMPER_PATCH_VERSION ${EMPER_PATCH_VERSION}

#define EMPER_GIT_COMMIT "${EMPER_GIT_COMMIT}"

#define EMPER_SHORT_VERSION "${EMPER_SHORT_VERSION}"

#define EMPER_FULL_VERSION "${EMPER_FULL_VERSION}"
EOF
}

printUsageAndExit() {
	echo "usage: ${0##*/} [-d} [-v} [--] [check <version>|clean]"
	exit 2
}

VERBOSE=false

while getopts :dv OPT; do
	case $OPT in
		d)
			set -x
			;;
		v)
			VERBOSE=true
			;;
		*)
			printUsageAndExit
	esac
done
shift $(( OPTIND - 1 ))
OPTIND=1


declare -r versionHeader="${BASEDIR}/emper/include/emper-version.h"

case $1 in
	check)
		checkVersionHeader "${2}"
		;;
	clean)
		rm -f "$versionHeader"
		;;
	*)
		printUsageAndExit
esac
